@@ -3,8 +3,13 @@ from .version import version as __version__
 
 __all__ = [
     "api",
+    "cli",
+    "collection",
     "exceptions",
     "market",
+    "nftmarket",
+    "nft",
+    "nfts",
     "rpc",
     "tokenobject",
     "tokens",
